use std::io::{stdout, Write};

use crossterm::{execute, queue};
use crossterm::cursor::{Hide, MoveTo, RestorePosition, SavePosition, Show};
use crossterm::style::Print;
use crossterm::terminal::{disable_raw_mode, enable_raw_mode, EnterAlternateScreen, LeaveAlternateScreen, SetSize, size};

use crate::VERSION;

pub const WIDTH: u16 = 160;
pub const HEIGHT: u16 = 40;
const HEADER_ROW: u16 = 0;
const FOOTER_ROW: u16 = HEIGHT - 1;
pub const TANK_ROW: u16 = FOOTER_ROW - 1;
pub const MIN_UFO_ROW: u16 = 1;
pub const MAX_UFO_ROW: u16 = TANK_ROW - 5;

pub struct Terminal {
    cols: u16,
    rows: u16,
}

pub fn setup() -> Terminal {
    let (cols, rows) = size()
        .expect("Failed to get terminal screen size");

    execute!(
        stdout(),
        EnterAlternateScreen,
        SetSize(WIDTH, HEIGHT),
        Hide
    ).expect("Failed to set up terminal screen");
    enable_raw_mode()
        .expect("Failed to enable raw terminal mode");

    write_header();
    write_footer();
    Terminal { cols, rows }
}

pub fn restore(terminal: Terminal) {
    disable_raw_mode()
        .expect("Failed to disable raw terminal mode");
    execute!(
        stdout(),
        SetSize(terminal.cols, terminal.rows),
        Show,
        LeaveAlternateScreen
    )
        .expect("Failed to tear down terminal screen");
}

fn write_header() {
    queue!(
        stdout(),
        MoveTo(0,HEADER_ROW),
        Print(center_text("UFO goes RUST! ** by fritteli, ufo@fritteli.ch ** Have fun!", None))
    ).expect("Failed to write to terminal");
    stdout().flush().expect("Failed to write to terminal");
}

fn write_footer() {
    queue!(
        stdout(),
        MoveTo(0,FOOTER_ROW),
        Print(center_text(format!("v{}",VERSION).as_str(), None))
    ).expect("Failed to write to terminal");
    stdout().flush().expect("Failed to write to terminal");
}

fn center_text(text: &str, filler: Option<&str>) -> String {
    let length = text.len();
    let surplus = (WIDTH as i16) - (length as i16) - 2;
    if surplus < 0 {
        return String::from(text);
    }
    let left: usize = (surplus as usize) / 2;
    let right: usize;
    if surplus % 2 == 0 {
        right = left;
    } else {
        right = left + 1;
    }
    let left_fill = filler.unwrap_or("=").repeat(left);
    let right_fill = filler.unwrap_or("=").repeat(right);
    format!("{} {} {}", left_fill, text, right_fill)
}

fn print_char_at(col: u16, row: u16, char: char) {
    execute!(
        stdout(),
        SavePosition,
        MoveTo(col, row),
        Print(char),
        RestorePosition
    )
        .expect("Failed to print char to terminal");
}

pub fn print_str_at(col: u16, row: u16, s: &str) {
    execute!(
        stdout(),
        SavePosition,
        MoveTo(col, row),
        Print(s),
        RestorePosition
    )
        .expect("Failed to print string to terminal");
}

pub fn clear_pos(col: u16, row: u16) {
    print_char_at(col, row, ' ');
}

pub fn clear_stretch(col: u16, row: u16, len: usize) {
    print_str_at(col, row, " ".repeat(len).as_str());
}
